﻿; ------------------------------------------------------------------------------------
; Stack Machine
; PB 4.5+ OS: Windows/Linux/MacOS Architecture: x86/x64
; Author: Alexander Aigner, cxAlex @ purebasic.fr/german
; V# 0.0.2
; ------------------------------------------------------------------------------------

EnableExplicit

; ------------------------------------------------------------------------------------
;{ Features:
;   - Einfach in eigene Projekte zu integrieren
;   - Einfach zu erweiternder Befehlssatz
;   - Sehr schnell
;   - Thread-Basiert
;   - Ausführung kann gestoppt und fortgesetzt werden
;   - Interrupt - Controller
;   - Frei definierbare Stackgröße
;}
; ------------------------------------------------------------------------------------

; ------------------------------------------------------------------------------------
;{ FAQ:
;

;}
; ------------------------------------------------------------------------------------

; ------------------------------------------------------------------------------------
;{ ToDo:
; SFP ausbauen, dokumentieren
; Speed optimieren
; Befehle erweitern und Doku immer mitschreiben!
; Timer
;}
; ------------------------------------------------------------------------------------

; ------------------------------------------------------------------------------------
;{ Changelog:
;
; - V# 0.0.2
; - Added: Genereller Interrupt-Controller
;
; - V# 0.0.1
; - Initial Release
;}
; ------------------------------------------------------------------------------------

; ------------------------------------------------------------------------------------
;{ Source - Quellen:


;}
; ------------------------------------------------------------------------------------

; CPU-Version
#SM_CPU_Version = 2

XIncludeFile "BinaryFormat.pbi" ; Binärformat
XIncludeFile "VirtualRam.pbi" ; Virtueller Arbeitsspeicher
XIncludeFile "VirtualRegister.pbi" ; Virtuelles Register

; Speicher-Verwaltung
Structure _SM_Memory
  Ram.VirtualRam ; Ram
  ; Register
  Akku.VirtualRegister
  Result.VirtualRegister
  CP.VirtualRegister
  SP.VirtualRegister
  BP.VirtualRegister
  OPWriter.VirtualRegister
  InterruptLevel.VirtualRegister
EndStructure

XIncludeFile "VirtualProcessor.pbi" ; Virtueller Processor
XIncludeFile "VirtualIO.pbi" ; Virtuelle IO - Verwaltung
XIncludeFile "VirtualInterruptController.pbi" ; Virtueller Interrupt - Controller

; Stack-Maschinen Daten
Structure _SM_Data
  *VTable
  Memory._SM_Memory ; Speicher
  Processor.VirtualProcessor ; Prozessor
  IO.VirtualIO ; I/O Einheit
  Interrupt.VirtualInterruptController ; Interrupt-Controller
  StackSize.l ; Stackgröße
  Running.a ; Läuft der Ausführungs-Thread?
  Stop.a ; Engine - Stop Flag
  Thread.i ; Ausführungs -Thread
EndStructure

; StackMachinen Interface
Interface StackMachine
  Reset() ; Zurücksetzen
  FlushMem() ; Speicher Fluschen
  AddCommand(OpCode.a, Param.l = #Null) ; OpCode schreiben
  Execute() ; Ausführung starten
  IsRunning() ; Läuft die Engine?
  Stop() ; Ausführung stoppen
  RegisterExternPortHandlers(ExportHandler, ImportHandler) ; Import/Export-Funktionen für User-IO Ports
  GetRegister.l(Register$) ; Inhalt eines Registers lesen
  Free() ; Freigeben
EndInterface

; Parser
XIncludeFile "Parser.pbi"

; Stack-Machine anlegen
Procedure StackMachine_New(Ram = 1024)
  Protected *SM._SM_Data = AllocateMemory(SizeOf(_SM_Data))
  Protected SM.StackMachine = *SM
  With *SM
    ; VTable:
    \VTable = ?_StackMachine_VTable
    ; Speicher anlegen:
    \Memory\Ram = _VRam_New(Ram*SizeOf(OpCode))
    \Memory\Akku = _VReg_New()
    \Memory\Result = _VReg_New()
    \Memory\CP = _VReg_New()
    \Memory\SP = _VReg_New()
    \Memory\BP = _VReg_New()
    \Memory\OPWriter = _VReg_New()
    \Memory\InterruptLevel = _VReg_New()
    \StackSize = Ram
    ; Processor erstellen:
    \Processor = _VProc_New(\Memory)
    ; I/O Einheit erstellen:
    \IO = _VIO_New(\Memory, *SM)
    ; Interrupt-Controller erstellen:
    \Interrupt = _VIntC_New(\Memory)
    ; Auf Anfang gehen:
    SM\Reset()
  EndWith
  
  ProcedureReturn *SM
EndProcedure

Procedure _StackMachine_Reset(*SM._SM_Data)
  With *SM
    ; Register setzen
    \Memory\Result\Set(#Null)
    \Memory\OPWriter\Set(#Null)
    \Memory\InterruptLevel\Set(#Null)
    \Memory\Akku\Set(#Null)
    \Memory\CP\Set(#Null)
    \Memory\BP\Set(\StackSize-1)
    \Memory\SP\Set(\Memory\BP\Get())
    
    \IO\IOPTU(0) ; IO Reseten
    \Interrupt\Reset() ; Interrupts Reseten
  EndWith
EndProcedure

Procedure _StackMachine_FlushMem(*SM._SM_Data)
  With *SM
    ; Register setzen
    _StackMachine_Reset(*SM)
    \Memory\Ram\Flush()
  EndWith
EndProcedure

Procedure _StackMachine_AddCommand(*SM._SM_Data, OpCode.a, Param.l)
  Protected OP.OpCode
  With *SM
    ; OpCode schreiben
    OP\OpCode = OpCode
    OP\Value = Param
    \Memory\Ram\SetOpCode(\Memory\OPWriter\Get(), @OP)
    \Memory\OPWriter\Inc()
  EndWith
EndProcedure

Procedure _StackMachine_ExecuteCMD(*SM._SM_Data)
  Protected *OP.OpCode
  With *SM
    ; OpCode laden...
    *OP = \Memory\Ram\GetOpCode(\Memory\CP\Get())
    Select *OP\OpCode ; Befehl ausführen...
      ; I/O Schnittstelle:  
      Case #OP_IN  : \IO\IN()
      Case #OP_OUT : \IO\OUT()
      Case #OP_IOPTU : \IO\IOPTU(*OP\Value)
      Case #OP_IOPTD : \IO\IOPTD(*OP\Value)
        ; Prozessor-Funktion:
        ; Load/Store Data
      Case #OP_LDAU : \Processor\LDAU(*OP\Value)
      Case #OP_LDAD : \Processor\LDAD(*OP\Value)
      Case #OP_STAD : \Processor\STAD(*OP\Value)
      Case #OP_LEA  : \Processor\LEA(*OP\Value)
        ; Arithmetic:
      Case #OP_ADDU  : \Processor\ADDU(*OP\Value)
      Case #OP_ADDD  : \Processor\ADDD(*OP\Value)
      Case #OP_SUBU : \Processor\SUBU(*OP\Value)
      Case #OP_SUBD : \Processor\SUBD(*OP\Value)
      Case #OP_MULU : \Processor\MULU(*OP\Value)
      Case #OP_MULD : \Processor\MULD(*OP\Value)
      Case #OP_DIVU : \Processor\DIVU(*OP\Value)
      Case #OP_DIVD : \Processor\DIVD(*OP\Value)
      Case #OP_MODU : \Processor\MODU(*OP\Value)
      Case #OP_MODD : \Processor\ModD(*OP\Value)
      Case #OP_ANDU : \Processor\ANDU(*OP\Value)
      Case #OP_ANDD : \Processor\ANDD(*OP\Value)
      Case #OP_ORU : \Processor\ORU(*OP\Value)
      Case #OP_ORD : \Processor\ORD(*OP\Value)
      Case #OP_XORU : \Processor\XORU(*OP\Value)
      Case #OP_XORD : \Processor\XORD(*OP\Value)
      Case #OP_NOT : \Processor\NOT(*OP\Value)
      Case #OP_SHLU : \Processor\SHLU(*OP\Value)
      Case #OP_SHLD : \Processor\SHLD(*OP\Value)
      Case #OP_SHRU : \Processor\SHRU(*OP\Value)
      Case #OP_SHRD : \Processor\SHRD(*OP\Value)
        ; Compare:
      Case #OP_CMPU : \Processor\CMPU(*OP\Value)
      Case #OP_CMPD : \Processor\CMPD(*OP\Value)
        ; Jumps:
      Case #OP_JMP : \Processor\JUM(*OP\Value)
      Case #OP_JP : \Processor\JP(*OP\Value)
      Case #OP_JNP : \Processor\JNP(*OP\Value)
      Case #OP_JN : \Processor\JN(*OP\Value)
      Case #OP_JNN : \Processor\JNN(*OP\Value)
      Case #OP_JZ : \Processor\JZ(*OP\Value)
      Case #OP_JNZ : \Processor\JNZ(*OP\Value)
      Case #OP_IF : \Processor\IF(*OP\Value)
      Case #OP_IFN : \Processor\IFN(*OP\Value)
        ; Procedure:
      Case #OP_CALL : \Processor\CAL(*OP\Value)
      Case #OP_RET : \Processor\RET(*OP\Value)
      Case #OP_RETN : \Processor\RETN(*OP\Value)
      Case #OP_NEWB : \Processor\NEWB(*OP\Value)
      Case #OP_OLDB : \Processor\OLDB(*OP\Value)
      Case #OP_RES : \Processor\RES(*OP\Value)
      Case #OP_Red : \Processor\Red(*OP\Value)
      Case #OP_PUSH : \Processor\PUSH(*OP\Value)
      Case #OP_POP : \Processor\POP(*OP\Value)
        ; Stack Command by Value:
      Case #OP_STL : \Processor\STL(*OP\Value)
      Case #OP_LDL : \Processor\LDL(*OP\Value)
      Case #OP_ADL : \Processor\ADL(*OP\Value)
      Case #OP_SBL : \Processor\SBL(*OP\Value)
      Case #OP_CPL : \Processor\CPL(*OP\Value)
        ; Stack Commands by Reference
      Case #OP_STLI : \Processor\STLI(*OP\Value)
      Case #OP_LDLI : \Processor\LDLI(*OP\Value)
      Case #OP_ADLI : \Processor\ADLI(*OP\Value)
      Case #OP_SBLI : \Processor\SBLI(*OP\Value)
      Case #OP_CPLI : \Processor\CPLI(*OP\Value)
    EndSelect
    
    ; Wenn wir nicht am Ende sind: CP erhöhen
    If *OP\OpCode <> #OP_Stop
      \Memory\CP\Inc()
    EndIf
    
    ProcedureReturn *OP\OpCode
  EndWith
EndProcedure

; Programm abarbeiten:
Procedure __StackMachine_Execute(*SM._SM_Data)
  With *SM
    ; Alles ausführen
    While _StackMachine_ExecuteCMD(*SM)
      \Interrupt\CheckInterrupts() ; auf Interrupts prüfen
      If \Stop ; Ausführung anhalten
        Break
      EndIf
    Wend
    \Running = #False
  EndWith
EndProcedure

; Abarabeitungs-Thread starten:
Procedure _StackMachine_Execute(*SM._SM_Data)
  With *SM
    If Not (\Running Or IsThread(\Thread))
      \Running = #True
      \Stop = #False
      \Thread = CreateThread(@__StackMachine_Execute(), *SM)  
    EndIf
  EndWith
EndProcedure

; Läuft der Ausführungs-Thread:
Procedure _StackMachine_IsRunning(*SM._SM_Data)
  With *SM
    ProcedureReturn \Running
  EndWith
EndProcedure

; Ausführung beenden:
Procedure _StackMachine_Stop(*SM._SM_Data)
  With *SM
    \Stop = #True
  EndWith
EndProcedure

; Externe Port - Handler setzen:
Procedure _StackMachine_RegisterExternPortHandlers(*SM._SM_Data, Export, Inport)
  With *SM
    \IO\RegisterExternPortHandlers(Export, Inport)
  EndWith
EndProcedure

; Register auslesen:
Procedure.l _StackMachine_GetRegister(*SM._SM_Data, Register.s)
  With *SM\Memory
    Select LCase(Register)
      Case "akku" : ProcedureReturn \Akku\Get()
      Case "result" : ProcedureReturn \Result\Get()
      Case "sp" : ProcedureReturn \SP\Get()
      Case "cp" : ProcedureReturn \CP\Get()
      Case "bp" : ProcedureReturn \bP\Get()
    EndSelect
  EndWith
EndProcedure

; StackMachine freigeben:
Procedure _StackMachine_Free(*SM._SM_Data)
  With *SM
    _StackMachine_Stop(*SM)
    ; Alles freigeben
    \Memory\Result\Free()
    \Memory\Akku\Free()
    \Memory\CP\Free()
    \Memory\BP\Free()
    \Memory\SP\Free()
    \Memory\OPWriter\Free()
    \Memory\InterruptLevel\Free()
    \Memory\Ram\Free()
    \Processor\Free()
    \IO\Free()
    \Interrupt\Free()
    ; Sich selbst freigeben
    FreeMemory(*SM)
  EndWith
EndProcedure


DataSection
  _StackMachine_VTable:
  Data.i @_StackMachine_Reset()
  Data.i @_StackMachine_FlushMem()
  Data.i @_StackMachine_AddCommand()
  Data.i @_StackMachine_Execute()
  Data.i @_StackMachine_IsRunning()
  Data.i @_StackMachine_Stop()
  Data.i @_StackMachine_RegisterExternPortHandlers()
  Data.i @_StackMachine_GetRegister()
  Data.i @_StackMachine_Free()
EndDataSection
; IDE Options = PureBasic 4.60 Beta 3 (Windows - x64)
; CursorPosition = 88
; FirstLine = 285
; Folding = g--
; EnableXP
; CPU = 1