﻿; ------------------------------------------------------------------------------------
; Stack Machine - VInterruptC
; PB 4.5+ OS: Windows/Linux/MacOS Architecture: x86/x64
; Author: Alexander Aigner, cxAlex @ purebasic.fr/german
; ------------------------------------------------------------------------------------

; Interrupt-Level über Ret/RetN dekrementieren und Interrupt-Confirm ausführen

; sollte reichen...
#VIntC_NbInterrupts = 511

; Interrupt - Controller für Timer-Interrupts, Hardware-Interrupts....
Interface VirtualInterruptController
  AddInterrupt(Interrupt.l, Address.l) ; Interrupt hinzufügen
  TriggerInterrupt(Interrupt.l, TriggerConfirmation = #False) ; Interrupt triggern
  CheckInterrupts() ; Getriggerte Interrupts ausführen
  Reset() ; zurücksetzen
  Free() 
EndInterface

; Daten eines Interrupts:
Structure _VIntC_Interrupt
  Address.l ; Addresse des Interrupts
  Triggered.a ; Wurde ausgelöst?
  TriggerConfirm.i ; Bestätigung fürs Triggern falls nötig.
EndStructure

; Interrupt-Verwaltung:
Structure _VIntC_Data
  *VTable
  *Mem._SM_Memory
  InterruptMutex.i
  InterruptCount.i
  InterruptLevel.i
  Array Interrupts._VIntC_Interrupt(#VIntC_NbInterrupts)
  Array InterruptConfirms.i(#VIntC_NbInterrupts)
EndStructure

; Neuer Interrupt-Controller
Procedure _VIntC_New(*Mem._SM_Memory)
  Protected *VIntC._VIntC_Data = AllocateMemory(SizeOf(_VIntC_Data))
  InitializeStructure(*VIntC, _VIntC_Data)
  
  With *VIntC
    \VTable = ?_VIntC_VTable
    \InterruptMutex = CreateMutex()
    \Mem = *Mem ; StackMachine Speicher
  EndWith
  
  ProcedureReturn *VIntC
EndProcedure

; Interrupt setzen:
Procedure _VIntC_AddInterrupt(*VIntC._VIntC_Data, Interrupt.l, Address.l)
  With *VIntC
    LockMutex(\InterruptMutex)
    
    ; Zähler anpassen:
    If \Interrupts(Interrupt)\Address And (Not Address)
      \InterruptCount - 1
    ElseIf (Not \Interrupts(Interrupt)\Address) And Address
      \InterruptCount + 1
    EndIf
    
    \Interrupts(Interrupt)\Address = Address
    UnlockMutex(\InterruptMutex)
  EndWith
EndProcedure

; Interrupt auslösen:
Procedure _VIntC_TriggerInterrupt(*VIntC._VIntC_Data, Interrupt.l, TriggerConfirmation = #False)
  With *VIntC
    LockMutex(\InterruptMutex)
    If \Interrupts(Interrupt)\Address
      \Interrupts(Interrupt)\Triggered = #True
      \Interrupts(Interrupt)\TriggerConfirm = TriggerConfirmation
    EndIf
    UnlockMutex(\InterruptMutex)
  EndWith
EndProcedure

; Ausgelöste Interrupts ausführen:
Procedure _VIntC_CheckInterrupts(*VIntC._VIntC_Data)
  Protected i
  With *VIntC
    ; Gibt es Interrupts?
    If \InterruptCount
      LockMutex(\InterruptMutex)
      ; Fertige Interrupts bestätigen: (Extern <> Intern)
      If \Mem\InterruptLevel\Get() <> \InterruptLevel
        \InterruptLevel - 1
        ; Bestätigen falls nötig:
        If \InterruptConfirms(\InterruptLevel)
          SignalSemaphore(\InterruptConfirms(\InterruptLevel))
          \InterruptConfirms(\InterruptLevel) = #Null
        EndIf
      EndIf
      ; Auslösen falls nötig
      For i = 0 To #VIntC_NbInterrupts
        If \Interrupts(i)\Triggered
          ; Call auf Interrup-Addresse:
          \Mem\Ram\Set(\Mem\SP\Get(), \Mem\CP\Get())
          \Mem\CP\Set(\Interrupts(i)\Address)
          \Mem\SP\Dec()
          ; Interrupt wurde ausgelöst
          \Interrupts(i)\Triggered = #False
          
          ; Auslösen bestätigen falls nötig
          If \Interrupts(i)\TriggerConfirm
            ; Confirm eintragen
            \InterruptConfirms(\InterruptLevel) = \Interrupts(i)\TriggerConfirm
            \Interrupts(i)\TriggerConfirm = #Null
          EndIf
          
          ; InterruptLevel erhöhen
          \InterruptLevel + 1 ; Intern erhöhen
          \Mem\InterruptLevel\Inc() ; Extern erhöhrn
          Break
        EndIf
      Next
      UnlockMutex(\InterruptMutex)
    EndIf
  EndWith
EndProcedure

; Reseten
Procedure _VIntC_Reset(*VIntC._VIntC_Data)
  Protected i
  With *VIntC
    LockMutex(\InterruptMutex)
    For i = 0 To #VIntC_NbInterrupts
      If \InterruptConfirms(i)
        SignalSemaphore(\InterruptConfirms(i))
        \InterruptConfirms(i) = #Null
      EndIf
      \Interrupts(i)\Address = #Null
      If \Interrupts(i)\TriggerConfirm
        SignalSemaphore(\Interrupts(i)\TriggerConfirm)
      EndIf
      \Interrupts(i)\TriggerConfirm = #Null
      \Interrupts(i)\Triggered = #Null
    Next
    \InterruptCount = #Null
    \InterruptLevel = #Null
    UnlockMutex(\InterruptMutex)
  EndWith
EndProcedure

; Interrupt-Controller freigeben:
Procedure _VIntC_Free(*VIntC._VIntC_Data)
  With *VIntC
    _VIntC_Reset(*VIntC)
    FreeMutex(\InterruptMutex)
    ClearStructure(*VIntC, _VIntC_Data)
    FreeMemory(*VIntC)
  EndWith
EndProcedure

; VTable:
DataSection
  _VIntC_VTable:
  Data.i @_VIntC_AddInterrupt()
  Data.i @_VIntC_TriggerInterrupt()
  Data.i @_VIntC_CheckInterrupts()
  Data.i @_VIntC_Reset()
  Data.i @_VIntC_Free()
EndDataSection

; IDE Options = PureBasic 4.60 Beta 3 (Windows - x64)
; CursorPosition = 114
; FirstLine = 79
; Folding = --
; EnableXP
; CPU = 1